/* $Header: /CVSROOT/diskimg/diskimg.c,v 1.7 2003/10/23 05:17:44 tino Exp $
 *
 * $Log: diskimg.c,v $
 * Revision 1.7  2003/10/23 05:17:44  tino
 * Log was not flushed immediately
 *
 * Revision 1.6  2003/10/23 04:57:25  tino
 * ditto
 *
 * Revision 1.5  2003/10/23 04:56:28  tino
 * some funny lines removed
 *
 * Revision 1.4  2003/10/23 04:55:22  tino
 * Forgot to set loops=0 on succesful read
 *
 * Revision 1.3  2003/10/23 03:57:05  tino
 * First working version
 *
 * Revision 1.2  2003/10/23 02:32:51  tino
 * First test of defect read management
 *
 * Revision 1.1  2003/10/23 01:19:55  tino
 * No defect skip mode yet.  First trial.
 */

#define	_GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <errno.h>

typedef	unsigned long long	POS;
#define	SECTSIZE		512llu
#define	SECTBLOCK		2048llu

static const char		*src;
static int			fd;
static FILE			*log;

static POS			start_sector, sectors, err_sectors, sectors_offset;

static void
ex(const char *s)
{
  fprintf(stderr, "\n");
  perror(s);
  exit(-1);
  abort();
  for(;;);
}

static void
openlog(const char *name)
{
  int	fd;

  if ((fd=open(name, O_CREAT|O_WRONLY|O_EXCL|O_SYNC|O_NOFOLLOW, 0777))<0)
    ex(name);
  if ((log=fdopen(fd, "a"))==NULL)
    ex("fdopen");
}

/* Show speed etc.
 * Second parameter is scale like in 2^(10*i)
 * 0 means byte (scale 1)
 * negative means milli(-1), micro, nano, pico, ato(-5)
 * positive means kilo(1), mega, giga, tera, peta, exa, zeta, yotta(8)
 */
static char *
scale6(POS p, int i)
{
  static const char	*ext="apnum KMGTPEZY";
  static char		buf[10];
  int			k;
  POS			div, rem;

  if (!p)
    i	= 0;
  i	+= 5;
  k	=  0;
  if (i<0)
    {
      k	= -i*10;
      i	= 0;
    }
  if (i>=strlen(ext))
    i	= strlen(ext)-1;
  for (; ext[++i] && (p>>k)>=10000llu; k+=10);
  i--;
  div	= 1llu<<k;
  rem	= p&(div-1ll);
  p	>>= k;

  if (p>=100 || (!p && !rem))
    sprintf(buf, "%4llu%cB", p, ext[i]);
  else if (p>=10)
    sprintf(buf, "%2llu.%llu%cB", p, rem*10llu/div, ext[i]);
  else if (p)
    sprintf(buf, "%llu.%02llu%cB", p, rem*100llu/div, ext[i]);
  else
    sprintf(buf, ".%03llu%cB", rem*1000llu/div, ext[i]);
  return buf;
}

static char *
scale(POS p)
{
  return scale6(p, 0);
}

static void
errrange(POS a, POS b)
{
  err_sectors++;
  if (a<b)
    {
      fprintf(stderr, "sector defect %llu-%llu (%s)\n", a, b, scale(SECTSIZE+SECTSIZE*(b-a)));
      if (log)
	fprintf(log, "%llu %llu\n", a, b);
      err_sectors+=b-a;
    }
  else
    {
      fprintf(stderr, "sector defect %llu (%s)\n", a, scale(SECTSIZE));
      if (log)
	fprintf(log, "%llu\n", a);
    }
  if (log)
    fflush(log);
}

static void
opensrc(const char *name)
{
  src	= name;
  if ((fd=open64(name, O_RDONLY|O_NDELAY|O_LARGEFILE))<0)
    ex(name);
}

static int	alarm_signal;

static void
alarmhandler(int sig)
{
#ifndef USE_SIGACTION
  signal(SIGALRM, alarmhandler);
#endif
  alarm_signal++;
  alarm(1);
}

static void
alarm_init(void)
{
#ifdef USE_SIGACTION
  struct sigaction sa;

  memset(&sa, 0, sizeof sa);
  sa.sa_handler	= alarmhandler;

  if (sigaction(SIGALRM, &sa, NULL))
    ex("sigaction");
#else
  if (signal(SIGALRM, alarmhandler))
    ex("signal");
#endif

  alarm(1);
}


static time_t	progress_begin;

static void
progress_bar(int op)
{
  time_t	now, diff;

  if (!alarm_signal && op>0)
    return;
  alarm_signal	= 0;

  time(&now);
  diff	= now-progress_begin;
  if (!diff)
    diff	= 1;
  fprintf(stderr, "%ld:%02ld:%02ld ", diff/3600, (diff/60)%60, diff%60);
  fprintf(stderr, "%s/s ", scale(sectors*SECTSIZE/(POS)diff));
  fprintf(stderr, "%s ", scale(sectors*SECTSIZE));
  fprintf(stderr, "%llu", start_sector+sectors);
  if (sectors_offset)
    fprintf(stderr, "@%llu", sectors_offset);
  fprintf(stderr, " ");
  fprintf(stderr, "%llu defects (%s) ", err_sectors, scale(err_sectors*SECTSIZE));
  fprintf(stderr, "state %d ", op);
  fprintf(stderr, " \r");
  fflush(stderr);
}

static void
progress_start(void)
{
  time(&progress_begin);
  alarm_signal	= 1;
  progress_bar(0);
  alarm_init();
}

static void
progress_stop(void)
{
  fprintf(stderr, "\n");
  alarm(0);
}

static void
writeout(const char *buf, int len)
{
  sectors_offset+= len;
  sectors	+= sectors_offset/SECTSIZE;
  sectors_offset%= SECTSIZE;

  while (len>0)
    {
      int	n;

      n	= write(1, buf, len);
      progress_bar(len);
      if (n<=0)
	{
	  if (!n)
	    ex("EOF on output. Disk full?");
	  if (errno==EINTR)
	    continue;
	  ex("output");
	}
      if (n>len)
	ex("weird");
      len	-= n;
      buf	+= n;
    }
}

static void
setpos(void)
{
  POS		want;
  off64_t	off, res;

  want	= (start_sector+sectors)*SECTSIZE+sectors_offset;
  off	= want;
  while ((res=lseek64(fd, off, SEEK_SET))==(off64_t)-1)
    if (errno!=EINTR)
      ex("lseek64");
  if (res!=want)
    ex("oops, seek did not get position");
}

static void
defective(void)
{
  POS	first=0, errs;
  int	cnt;

#if 0
  fprintf(stderr, "defective search at sector %llu\n", sectors);
#endif
  errs	= 0;
  for (cnt=SECTBLOCK; cnt>0; )
    {
      char	buf[SECTSIZE];
      int	n;
      
      setpos();
      n	= read(fd, buf, SECTSIZE);
      progress_bar(-(int)errs);
      if (n==SECTSIZE)
	{
	  writeout(buf, n);
	  cnt--;
	  if (!errs)
	    continue;

	  errrange(first, first+ --errs);
	  break;
	}
      if (n>0)
	ex("unexpected partial read");
      if (!n)
	ex("unexpected EOF");
      switch (errno)
	{
	case EINTR:
	  continue;

	case EAGAIN:
	  sleep(1);
	  continue;

	case EIO:
	  break;

	default:
	  ex("read defective");
	}
      if (!errs++)
	first	= start_sector+sectors;
      memset(buf, 0, sizeof buf);
      {
	time_t	now;

	time(&now);
	sprintf(buf, "DISKIMG DEFECTIVE SECTOR %llu - %s - %s", start_sector+sectors, src, ctime(&now));
      }
      writeout(buf, SECTSIZE);
    }
}

#if SECTSIZE&(SECTSIZE-1)
#error "SECTSIZE must be power of 2"
#endif

static void
copyloop(void)
{
  int	loops, defect;

  defect	= 0;
  for (loops=0;loops<10000;loops++)
    {
      char	buf[SECTSIZE*SECTBLOCK];
      int	n;

      if (defect)
	{
	  defect=0;
	  defective();
	  loops=0;
	  continue;
	}
      if ((loops&0xf)==9)
	sleep(1);
      n	= read(fd, buf, sizeof buf);
      progress_bar(n);
      if (n<=0)
	{
	  if (loops==8)		/* try something clever ..	*/
	    sync();
	  if (!n)
	    {
	      /* Try really hard for EOF	*/
	      if (loops>15)
		break;
	      continue;
	    }
	  switch (errno)
	    {
	    default:
	      ex("read");

	    case EINTR:
	    case EAGAIN:
	      break;

	    case EIO:
	      defect	= 1;
	      break;
	    }
	  continue;
	}
      if (n%SECTSIZE)
	{
	  fprintf(stderr, "partial sector size %llu\n", n%SECTSIZE);
	  defect	= 1;
	  n&=~(SECTSIZE-1);
	}
      writeout(buf, n);
      loops=0;
    }
  ex("too many loops");
}

static POS
getoffset(const char *s)
{
  POS	off;
  char	*end;

  off	= strtoull(s, &end, 0);
  switch (*end)
    {
    case 0:
      return off;

    case 'b':
    case 'B':
      if (off%SECTSIZE)
	ex("you must give full sector counts");
      off/=SECTSIZE;
      break;
    
    case 'g':
    case 'G':
      off*=1024llu;
    case 'm':
    case 'M':
      off*=1024llu;
    case 'k':
    case 'K':
      off+=off;
      break;
    }
  if (*++end)
    ex("cannot parse start sector number");

  return off;
}

int
main(int argc, char **argv)
{
  if (argc<2 || argc>4)
    {
      fprintf(stderr,
	      "Usage: %s drive [sector [log]] 2>progress >image\n"
	      "\tRead a (possibly partly defective) drive,\n"
	      "\tstart with a particuliar sector (or offset if BKMG present),\n"
	      "\twrite the skipped (unreadable) sectors to a log,\n"
	      "\tprint progress and infos to stderr\n"
	      "\tand send image to stdout for further processing.\n"
	      "\tHint: diskimg /dev/hda1 1G | lzop | split -b 1024m hda1.gz.\n"
	      , argv[0]);
      return 1;
    }

  if (argc>2)
    start_sector	= getoffset(argv[2]);

  if (argc>3)
    openlog(argv[3]);

  opensrc(argv[1]);
  setpos();

  progress_start();
  copyloop();
  progress_stop();

  return 0;
}
